import joblib
import matplotlib.pyplot as plt
import pymc3 as pm
import seaborn as sns
from sklearn.base import BaseEstimator

from pymc3_models.exc import PyMC3ModelsError


class BayesianModel(BaseEstimator):
    """
    Bayesian model base class
    """
    def __init__(self):
        self.cached_model = None
        self.inference_type = None
        self.num_pred = None
        self.shared_vars = None
        self.summary = None
        self.trace = None

    def create_model(self):
        raise NotImplementedError

    def _set_shared_vars(self, shared_vars):
        """
        Sets theano shared variables for the PyMC3 model.
        """
        for key in shared_vars.keys():
            self.shared_vars[key].set_value(shared_vars[key])

    def _inference(self, inference_type='advi', inference_args=None):
        """
        Calls internal methods for two types of inferences. Raises an error if the inference_type is not supported.

        Parameters
        ==========
        inference_type : string, specifies which inference method to call. Defaults to 'advi'. Currently, only 'advi' and 'nuts' are supported

        inference_args : dict, arguments to be passed to the inference methods. Check the PyMC3 docs to see what is permitted. Defaults to None.
        """
        if inference_type == 'advi':
            self._advi_inference(inference_args)
        elif inference_type == 'nuts':
            self._nuts_inference(inference_args)
        else:
            raise PyMC3ModelsError('{} is not a supported type of inference'.format(inference_type))

    def _advi_inference(self, inference_args):
        """
        Runs variational ADVI and then samples from those results.

        Parameters
        ----------
        inference_args : dict, arguments to be passed to the PyMC3 fit method. See PyMC3 doc for permissible values.
        """
        with self.cached_model:
            inference = pm.ADVI()
            approx = pm.fit(method=inference, **inference_args)

        self.approx = approx
        self.trace = approx.sample(draws=10000)
        self.summary = pm.df_summary(self.trace)
        self.advi_hist = inference.hist

    def _nuts_inference(self, inference_args):
        """
        Runs NUTS inference.

        Parameters
        ----------
        inference_args : dict, arguments to be passed to the PyMC3 sample method. See PyMC3 doc for permissible values.
        """
        with self.cached_model:
            step = pm.NUTS()
            nuts_trace = pm.sample(step=step, **inference_args)

        self.trace = nuts_trace
        self.summary = pm.df_summary(self.trace)

    def _set_default_inference_args(self):
        """
        Set default values for inference arguments if none are provided, dependent on inference type.

        ADVI
        -----
        callbacks : list containing a parameter stopping check.

        n : number of iterations for ADVI fit, defaults to 200000

        NUTS
        -----
        draws : the number of samples to draw, defaults to 2000
        """
        if self.inference_type == 'advi':
            inference_args = {
                'n': 200000,
                'callbacks': [pm.callbacks.CheckParametersConvergence()]
            }
        elif self.inference_type == 'nuts':
            inference_args = {
                'draws': 2000
            }
        else:
            inference_args = None

        return inference_args

    def fit(self):
        raise NotImplementedError

    def predict(self):
        raise NotImplementedError

    def score(self):
        raise NotImplementedError

    def save(self, file_prefix, custom_params=None):
        """
        Saves the trace and custom params to files with the given file_prefix.

        Parameters
        ----------
        file_prefix : str, path and prefix used to identify where to save the trace for this model.
            Ex: given file_prefix = "path/to/file/"
            This will attempt to save to "path/to/file/trace.pickle"

        custom_params : Dictionary of custom parameters to save. Defaults to None
        """
        fileObject = open(file_prefix + 'trace.pickle', 'wb')
        joblib.dump(self.trace, fileObject)
        fileObject.close()

        if custom_params:
            fileObject = open(file_prefix + 'params.pickle', 'wb')
            joblib.dump(custom_params, fileObject)
            fileObject.close()

    def load(self, file_prefix, load_custom_params=False):
        """
        Loads a saved version of the trace, and custom param files with the given file_prefix.

        Parameters
        ----------
        file_prefix : str, path and prefix used to identify where to load the saved trace for this model.
            Ex: given file_prefix = "path/to/file/"
            This will attempt to load "path/to/file/trace.pickle"

        load_custom_params : Boolean flag to indicate whether custom parameters should be loaded. Defaults to False.

        Returns
        ----------
        custom_params : Dictionary of custom parameters
        """
        self.trace = joblib.load(file_prefix + 'trace.pickle')

        custom_params = None
        if load_custom_params:
            custom_params = joblib.load(file_prefix + 'params.pickle')

        return custom_params

    def plot_elbo(self):
        """
        Plot the ELBO values after running ADVI minibatch.
        """
        if self.inference_type != 'advi':
            raise PyMC3ModelsError(
                'This method should only be called after calling fit with ADVI minibatch.'
            )

        sns.set_style("white")
        plt.plot(-self.advi_hist)
        plt.ylabel('ELBO')
        plt.xlabel('iteration')
        sns.despine()



import shutil
import tempfile
import unittest

import numpy as np
from pymc3 import summary
from sklearn.model_selection import train_test_split

from pymc3_models.exc import PyMC3ModelsError
from pymc3_models import HierarchicalLogisticRegression


class HierarchicalLogisticRegressionTestCase(unittest.TestCase):
    def setUp(self):
        def numpy_invlogit(x):
            return 1 / (1 + np.exp(-x))

        self.num_cats = 3
        self.num_pred = 1
        self.num_samples_per_cat = 100000

        self.alphas = np.random.randn(self.num_cats)
        self.betas = np.random.randn(self.num_cats, self.num_pred)
        #TODO: make this more efficient; right now, it's very explicit so I understand it.
        x_a = np.random.randn(self.num_samples_per_cat, self.num_pred)
        y_a = np.random.binomial(1, numpy_invlogit(self.alphas[0] + np.sum(self.betas[0] * x_a, 1)))
        x_b = np.random.randn(self.num_samples_per_cat, self.num_pred)
        y_b = np.random.binomial(1, numpy_invlogit(self.alphas[1] + np.sum(self.betas[1] * x_b, 1)))
        x_c = np.random.randn(self.num_samples_per_cat, self.num_pred)
        y_c = np.random.binomial(1, numpy_invlogit(self.alphas[2] + np.sum(self.betas[2] * x_c, 1)))

        X = np.concatenate([x_a, x_b, x_c])
        Y = np.concatenate([y_a, y_b, y_c])
        cats = np.concatenate([
            np.zeros(self.num_samples_per_cat, dtype=np.int),
            np.ones(self.num_samples_per_cat, dtype=np.int),
            2*np.ones(self.num_samples_per_cat, dtype=np.int)
        ])

        self.X_train, self.X_test, self.cat_train, self.cat_test, self.Y_train, self.Y_test = train_test_split(
            X, cats, Y, test_size=0.4
        )

        self.test_HLR = HierarchicalLogisticRegression()

        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)


class HierarchicalLogisticRegressionFitTestCase(HierarchicalLogisticRegressionTestCase):
    def test_fit_returns_correct_model(self):
        # Note: print is here so PyMC3 output won't overwrite the test name
        print('')
        self.test_HLR.fit(self.X_train, self.Y_train, self.cat_train, minibatch_size=2000)

        self.assertEqual(self.num_cats, self.test_HLR.num_cats)
        self.assertEqual(self.num_pred, self.test_HLR.num_pred)

        #TODO: Figure out best way to test
        #np.testing.assert_almost_equal(self.alphas, self.test_HLR.trace['alphas'].mean(), decimal=1)
        #np.testing.assert_almost_equal(self.betas, self.test_HLR.trace['betas'].mean(), decimal=1)

        # For now, just check that the estimated parameters have the correct signs
        np.testing.assert_equal(
            np.sign(self.alphas),
            np.sign(self.test_HLR.trace['alpha'].mean(axis=0))
        )
        np.testing.assert_equal(
            np.sign(self.betas),
            np.sign(self.test_HLR.trace['beta'].mean(axis=0))
        )


class HierarchicalLogisticRegressionPredictProbaTestCase(HierarchicalLogisticRegressionTestCase):
    def test_predict_proba_returns_probabilities(self):
        print('')
        self.test_HLR.fit(self.X_train, self.Y_train, self.cat_train, minibatch_size=2000)
        probs = self.test_HLR.predict_proba(self.X_test, self.cat_test)
        self.assertEqual(probs.shape, self.Y_test.shape)

    def test_predict_proba_returns_probabilities_and_std(self):
        print('')
        self.test_HLR.fit(self.X_train, self.Y_train, self.cat_train, minibatch_size=2000)
        probs, stds = self.test_HLR.predict_proba(self.X_test, self.cat_test, return_std=True)
        self.assertEqual(probs.shape, self.Y_test.shape)
        self.assertEqual(stds.shape, self.Y_test.shape)

    def test_predict_proba_raises_error_if_not_fit(self):
        with self.assertRaises(PyMC3ModelsError) as no_fit_error:
            test_HLR = HierarchicalLogisticRegression()
            test_HLR.predict_proba(self.X_train, self.cat_train)

        expected = 'Run fit on the model before predict.'
        self.assertEqual(str(no_fit_error.exception), expected)


class HierarchicalLogisticRegressionPredictTestCase(HierarchicalLogisticRegressionTestCase):
    def test_predict_returns_predictions(self):
        print('')
        self.test_HLR.fit(self.X_train, self.Y_train, self.cat_train, minibatch_size=2000)
        preds = self.test_HLR.predict(self.X_test, self.cat_test)
        self.assertEqual(preds.shape, self.Y_test.shape)


class HierarchicalLogisticRegressionScoreTestCase(HierarchicalLogisticRegressionTestCase):
    def test_score_scores(self):
        print('')
        self.test_HLR.fit(self.X_train, self.Y_train, self.cat_train, minibatch_size=2000)
        score = self.test_HLR.score(self.X_test, self.Y_test, self.cat_test)
        naive_score = np.mean(self.Y_test)
        self.assertGreaterEqual(score, naive_score)


class HierarchicalLogisticRegressionSaveandLoadTestCase(HierarchicalLogisticRegressionTestCase):
    def test_save_and_load_work_correctly(self):
        print('')
        self.test_HLR.fit(self.X_train, self.Y_train, self.cat_train, minibatch_size=2000)
        probs1 = self.test_HLR.predict_proba(self.X_test, self.cat_test)
        self.test_HLR.save(self.test_dir)

        HLR2 = HierarchicalLogisticRegression()

        HLR2.load(self.test_dir)

        self.assertEqual(self.test_HLR.num_cats, HLR2.num_cats)
        self.assertEqual(self.test_HLR.num_pred, HLR2.num_pred)
        self.assertEqual(self.test_HLR.num_training_samples, HLR2.num_training_samples)
        self.assertEqual(summary(self.test_HLR.trace), summary(HLR2.trace))

        probs2 = HLR2.predict_proba(self.X_test, self.cat_test)

        np.testing.assert_almost_equal(probs2, probs1, decimal=1)



import unittest

from pymc3_models.models import BayesianModel


class BayesianModelTestCase(unittest.TestCase):
    def test_create_model_raises_not_implemented_error(self):
        with self.assertRaises(NotImplementedError):
            BM = BayesianModel()
            BM.create_model()

    def test_fit_raises_not_implemented_error(self):
        with self.assertRaises(NotImplementedError):
            BM = BayesianModel()
            BM.fit()

    def test_predict_raises_not_implemented_error(self):
        with self.assertRaises(NotImplementedError):
            BM = BayesianModel()
            BM.predict()

    def test_score_raises_not_implemented_error(self):
        with self.assertRaises(NotImplementedError):
            BM = BayesianModel()
            BM.score()



from setuptools import setup, find_packages

with open('README.rst') as r:
    readme = r.read()

with open('AUTHORS.txt') as a:
    # reSt-ify the authors list
    authors = ''
    for author in a.read().split('\n'):
        authors += '| '+author+'\n'

with open('LICENSE') as l:
    license = l.read()


setup(
    name='pymc3_models',
    version='1.0.1',
    description='Custom PyMC3 models built on top of the scikit-learn API',
    long_description=readme,
    author='Nicole Carlson',
    author_email='nicole@parsingscience.com',
    url='https://github.com/parsing-science/pymc3_models',
    license=license,
    packages=find_packages(),
    package_data={'docs': ['*']},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'joblib',
        'matplotlib',
        'numpy',
        'pandas>=0.19',
        'pymc3>=3.2',
        'scipy',
        'seaborn',
        'sklearn'
    ],
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4'
    ]
)
